from typing import Generic, Protocol, TypeVar

import torch
import torch.nn.functional as F


def convert_to_xyxy(boxes: torch.Tensor) -> torch.Tensor:
    return torch.cat(
        (
            boxes[..., :2] - boxes[..., 2:] / 2,
            boxes[..., :2] + boxes[..., 2:] / 2,
        ),
        dim=-1,
    )


def intersect(box_a: torch.Tensor, box_b: torch.Tensor) -> torch.Tensor:
    # [batch, n_obj, n_anchors, 2] ~
    max_xy = torch.min(box_a[..., 2:], box_b[..., 2:])
    min_xy = torch.max(box_a[..., :2], box_b[..., :2])

    inter = torch.clamp(max_xy - min_xy, min=0)
    # [batch, n_obj, n_anchors] ~
    return inter[..., 0] * inter[..., 1]


def iou(box_a: torch.Tensor, box_b: torch.Tensor) -> torch.Tensor:
    # [batch, n_obj, n_anchors] ~
    inter = intersect(box_a, box_b)

    # [batch, n_obj] ~
    area_a = (box_a[..., 2] - box_a[..., 0]) * (box_a[..., 3] - box_a[..., 1])
    # [batch, n_anchors] ~
    area_b = (box_b[..., 2] - box_b[..., 0]) * (box_b[..., 3] - box_b[..., 1])
    union = area_a + area_b - inter

    # [batch, n_obj, n_anchors] ~
    return inter / union


def match_boxes(
    boxes: torch.Tensor,  # [n_obj, 4]
    priors: torch.Tensor,  # [n_anchors, 4]
    overlap_threshold: float,
) -> torch.Tensor:  # returns a tensor of shape [n_anchors, n_obj] ~
    n_anchors = priors.shape[0]
    n_obj = boxes.shape[0]

    # Compute IoU overlaps: [n_obj, n_anchors]
    overlaps = iou(boxes[:, None], convert_to_xyxy(priors))

    # For each ground truth box, find the best matching prior (anchor)
    best_prior_overlap, best_prior_idx = overlaps.max(dim=1, keepdim=True)
    valid_gt_idx = best_prior_overlap[:, 0] >= 0.2

    # For each prior (anchor), find the best matching ground truth box
    best_truth_overlap, best_truth_idx = overlaps.max(dim=0, keepdim=True)
    best_truth_overlap = best_truth_overlap.squeeze(0)
    best_truth_idx = best_truth_idx.squeeze(0)
    best_prior_idx = best_prior_idx.squeeze(1)

    # Even if no good matches, return a zero-filled match matrix
    matching_table = torch.zeros(
        (n_anchors, n_obj), dtype=torch.bool, device=boxes.device
    )
    if valid_gt_idx.sum() == 0:
        return matching_table

    # Force match: assign each valid GT to its best matching prior
    best_truth_overlap.index_fill_(0, best_prior_idx[valid_gt_idx], 2)
    best_truth_idx[best_prior_idx] = torch.arange(
        best_prior_idx.shape[0], device=best_prior_idx.device
    )

    # Mark anchors as positive if IoU exceeds threshold
    valid_anchors = best_truth_overlap >= overlap_threshold
    matching_table[valid_anchors, best_truth_idx[valid_anchors]] = 1
    return matching_table


def mine_negatives(
    label: torch.Tensor,
    pred: torch.Tensor,
    negpos_ratio: int,
    positive: torch.Tensor,
) -> torch.Tensor:
    batch_size, num_anchors, _ = positive.shape
    pos_batch, pos_anchor, pos_obj = torch.where(positive)
    labels = torch.zeros_like(pred[:, :, 0], dtype=torch.long)
    labels[pos_batch, pos_anchor] = label[pos_batch, pos_obj].squeeze()
    loss = F.cross_entropy(
        pred.view(-1, pred.shape[-1]), labels.view(-1), reduction="none"
    ).view(batch_size, num_anchors)
    loss[pos_batch, pos_anchor] = 0
    _, loss_sorted_idx = loss.sort(dim=1, descending=True)
    _, rank = loss_sorted_idx.sort(dim=1)
    num_pos = positive.sum(dim=(1, 2), dtype=torch.long).unsqueeze(1)
    num_neg = torch.clamp(negpos_ratio * num_pos, max=num_anchors - 1)
    return rank < num_neg.expand_as(rank)


def iterative_mathing(
    classes: torch.Tensor,  # [batch_size, n_anchors, n_classes]
    boxes: torch.Tensor,  # [batch_size, n_obj, 4]
    priors: torch.Tensor,  # [n_anchors, 4]
    confidences: torch.Tensor,  # [batch_size, n_anchors, n_classes]
    negpos_ratio: int,
    overalp: float,
) -> tuple[torch.Tensor, torch.Tensor]:
    positives = torch.stack([match_boxes(b, priors, overalp) for b in boxes])
    negatives = mine_negatives(classes, confidences, negpos_ratio, positives)
    return positives, negatives


T = TypeVar("T")


class HasBoxesAndClasses(Protocol, Generic[T]):
    boxes: T
    classes: T


def match(
    y_pred: HasBoxesAndClasses[torch.Tensor],
    y_true: HasBoxesAndClasses[torch.Tensor],
    anchors: torch.Tensor,
    negpos_ratio: int,
    overalp: float,
) -> tuple[torch.Tensor, torch.Tensor]:
    return iterative_mathing(
        y_true.classes,
        y_true.boxes,
        anchors,
        confidences=y_pred.classes,
        negpos_ratio=negpos_ratio,
        overalp=overalp,
    )



from itertools import product
from math import ceil

import torch


def anchors(
    min_sizes: list[list[int]],
    steps: list[int],
    clip: bool,
    resolution: tuple[int, int],  # h, w resolution
) -> torch.Tensor:
    feature_maps = [
        [ceil(resolution[0] / step), ceil(resolution[1] / step)]
        for step in steps
    ]

    anchors: list[float] = []
    for k, f in enumerate(feature_maps):
        t_min_sizes = min_sizes[k]
        for i, j in product(range(f[0]), range(f[1])):
            for min_size in t_min_sizes:
                s_kx = min_size / resolution[1]
                s_ky = min_size / resolution[0]
                dense_cx = [x * steps[k] / resolution[1] for x in [j + 0.5]]
                dense_cy = [y * steps[k] / resolution[0] for y in [i + 0.5]]
                for cy, cx in product(dense_cy, dense_cx):
                    anchors += [cx, cy, s_kx, s_ky]

    # back to torch land
    output = torch.Tensor(anchors).view(-1, 4)
    if clip:
        output.clamp_(max=1, min=0)
    return output



import functools
from dataclasses import dataclass
from typing import Callable, Optional, Union

import torch

LossFunctionyType = Union[
    torch.nn.Module,
    Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
]


@dataclass
class WeightedLoss:
    loss: Optional[LossFunctionyType]
    weight: float = 1.0
    dec_pred: Callable = lambda x, _: x
    enc_pred: Callable = lambda x, _: x
    enc_true: Callable = lambda x, _: x
    needs_negatives: bool = False

    def __call__(self, y_pred, y_true, anchors):
        y_pred_encoded = self.enc_pred(y_pred, anchors)
        y_true_encoded = self.enc_true(y_true, anchors)
        return self.weight * self.loss(y_pred_encoded, y_true_encoded)


def masked_loss(loss_function: LossFunctionyType) -> LossFunctionyType:
    @functools.wraps(loss_function)
    def f(pred: torch.Tensor, data: torch.Tensor) -> torch.Tensor:
        mask = ~torch.isnan(data)
        data_masked = data[mask]
        pred_masked = pred[mask]
        loss = loss_function(data_masked, pred_masked)
        if data_masked.numel() == 0:
            loss = torch.nan_to_num(loss, 0)
        return loss / max(data_masked.shape[0], 1)

    return f


def retina_confidence_loss(
    y_pred: torch.Tensor,
    y_true: torch.Tensor,
) -> tuple[torch.Tensor]:
    n_pos = (y_true > 0).sum()
    loss = torch.nn.functional.cross_entropy(
        y_pred,
        y_true.view(-1),
        reduction="sum",
    )
    return loss / n_pos



from dataclasses import fields, is_dataclass
from functools import partial
from typing import Callable, Generic, Protocol, Tuple, TypeVar

import torch
from torch import nn

from detectinhos.matching import match
from detectinhos.sublosses import WeightedLoss


def select(
    y_pred: torch.Tensor,
    y_true: torch.Tensor,
    anchors: torch.Tensor,
    use_negatives: bool,
    positives: torch.Tensor,
    negatives: torch.Tensor,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    b_pos, a_pos, o_pos = torch.where(positives)
    pred_pos = y_pred[b_pos, a_pos]
    true_pos = y_true[b_pos, o_pos]
    anch_pos = anchors[a_pos]

    if not use_negatives:
        return pred_pos, true_pos, anch_pos

    b_neg, a_neg = torch.where(negatives)
    pred_neg = y_pred[b_neg, a_neg]
    true_neg = torch.zeros_like(pred_neg[:, 0], dtype=torch.long)
    anch_neg = anchors[a_neg]

    pred_all = torch.cat([pred_pos, pred_neg], dim=0)
    true_all = torch.cat([true_pos.view(-1), true_neg], dim=0).long()
    anch_all = torch.cat([anch_pos, anch_neg], dim=0)
    return pred_all, true_all, anch_all


T = TypeVar("T")
LossContainer = TypeVar(
    "LossContainer",
    bound="HasBoxesAndClasses[WeightedLoss]",
)


class HasBoxesAndClasses(Protocol, Generic[T]):
    scores: T
    boxes: T
    classes: T

    @classmethod
    def is_dataclass(cls) -> bool: ...


Matching = Callable[
    [
        HasBoxesAndClasses[torch.Tensor],
        HasBoxesAndClasses[torch.Tensor],
        torch.Tensor,
    ],
    Tuple[torch.Tensor, torch.Tensor],
]


class DetectionLoss(Generic[LossContainer], nn.Module):
    def __init__(
        self,
        priors: torch.Tensor,
        sublosses: LossContainer,
        match: Matching = partial(
            match,
            negpos_ratio=7,
            overalp=0.35,
        ),
    ) -> None:
        super().__init__()
        if not is_dataclass(sublosses):
            raise TypeError("sublosses must be a dataclass instance")
        self.sublosses = sublosses
        self.match = match
        self.register_buffer("priors", priors)

    def forward(
        self,
        y_pred: HasBoxesAndClasses[torch.Tensor],
        y_true: HasBoxesAndClasses[torch.Tensor],
    ) -> dict[str, torch.Tensor]:
        positives, negatives = self.match(
            y_pred,
            y_true,
            self.priors,
        )

        losses = {}
        for field in fields(self.sublosses):
            name = field.name
            subloss: WeightedLoss = getattr(self.sublosses, name)
            if subloss.loss is None:
                continue
            y_pred_, y_true_, anchor_ = select(
                getattr(y_pred, name),
                getattr(y_true, name),
                self.priors,
                use_negatives=subloss.needs_negatives,
                positives=positives,
                negatives=negatives,
            )
            losses[name] = subloss(y_pred_, y_true_, anchor_)

        losses["loss"] = torch.stack(tuple(losses.values())).sum()
        return losses



import json
import pathlib

import cv2
import numpy as np
import pytest
import torch

from detectinhos.anchors import anchors
from detectinhos.encode import encode


@pytest.fixture
def true():
    # [xmin, ymin, xmax, ymax, class_id, difficult, crowd] ~
    # Convert to relative coordinates (default resolution is 640x480)
    width, height = 640, 480
    data = np.array(
        [
            [439, 157, 556, 241, 0, 0, 0],
            [437, 246, 518, 351, 0, 0, 0],
            [515, 306, 595, 375, 0, 0, 0],
            [407, 386, 531, 476, 0, 0, 0],
            [544, 419, 621, 476, 0, 0, 0],
            [609, 297, 636, 392, 0, 0, 0],
        ],
        dtype=np.float32,
    )
    data[:, [0, 2]] /= width  # x-coordinates
    data[:, [1, 3]] /= height  # y-coordinates
    return data


@pytest.fixture
def pred():
    # [xmin, ymin, xmax, ymax, class_id, confidence] ~
    # Convert to relative coordinates (default resolution is 640x480)
    width, height = 640, 480
    data = np.array(
        [
            [429, 219, 528, 247, 0, 0.460851],
            [433, 260, 506, 336, 0, 0.269833],
            [518, 314, 603, 369, 0, 0.462608],
            [592, 310, 634, 388, 0, 0.298196],
            [403, 384, 517, 461, 0, 0.382881],
            [405, 429, 519, 470, 0, 0.369369],
            [433, 272, 499, 341, 0, 0.272826],
            [413, 390, 515, 459, 0, 0.619459],
        ],
        dtype=np.float32,
    )
    data[:, [0, 2]] /= width  # x-coordinates
    data[:, [1, 3]] /= height  # y-coordinates
    return data


@pytest.fixture
def annotations(tmp_path, image, true) -> pathlib.Path:
    fname = str(tmp_path / "image.png")
    cv2.imwrite(fname, image)

    # Use true labels to populate the annotations
    example = [
        {
            "file_name": fname,
            "annotations": [
                {
                    "label": "apple",
                    "bbox": true[i][:4].tolist(),
                }
                for i in range(true.shape[0])
            ],
        },
    ]
    ofile = tmp_path / "annotations.json"
    with open(ofile, "w") as f:
        json.dump(example, f, indent=2)
    return ofile


@pytest.fixture
def boxes_true(true) -> torch.Tensor:
    return torch.Tensor(true[:, :4]).unsqueeze(0)


@pytest.fixture
def classes_true(true) -> torch.Tensor:
    return torch.Tensor(true[:, 4]).unsqueeze(0) + 1


@pytest.fixture
def boxes_pred(pred, sample_anchors) -> torch.Tensor:
    total = torch.zeros((sample_anchors.shape[0], 4), dtype=torch.float32)
    total[: pred.shape[0]] = torch.Tensor(pred[:, :4])
    return encode(
        total,
        sample_anchors,
        variances=[0.1, 0.2],
    ).unsqueeze(0)


@pytest.fixture
def classes_pred(pred, sample_anchors) -> torch.Tensor:
    total = torch.zeros((sample_anchors.shape[0], 2), dtype=torch.float32)
    # Everything is background
    total[:, 0] = 1.0
    # Except for the predictions
    total[: pred.shape[0]] = torch.Tensor(pred[:, 4:6])
    total[: pred.shape[0], 0] = 1 - total[: pred.shape[0], 1]
    # Convert probabilities to logits
    eps = 1e-7
    total = torch.clamp(total, eps, 1.0)  # ensure log safety
    logits = torch.log(total)
    return logits.unsqueeze(0)


@pytest.fixture
def image(resolution: tuple[int, int]) -> np.ndarray:
    return np.random.randint(0, 255, resolution + (3,), dtype=np.uint8)


@pytest.fixture
def sample_anchors(image) -> torch.Tensor:
    return anchors(
        min_sizes=[[16, 32], [64, 128], [256, 512]],
        steps=[8, 16, 32],
        clip=False,
        resolution=image.shape[:2],
    )



import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Generic, TypeVar

from dacite import Config, from_dict
from dataclasses_json import dataclass_json

RelativeXYXY = tuple[float, float, float, float]


@dataclass_json
@dataclass
class Annotation:
    bbox: RelativeXYXY
    label: str
    score: float = float("nan")


T = TypeVar("T", bound=Annotation)


@dataclass
class Sample(Generic[T]):
    file_name: str
    annotations: list[T]


def to_sample(
    entry: dict[str, Any],
    sample_type: type[Sample[T]],
) -> Sample[T]:
    return from_dict(
        data_class=sample_type,
        data=entry,
        config=Config(cast=[tuple]),
    )


def read_dataset(
    path: Path | str,
    sample_type: type[Sample[T]] = Sample[T],
) -> list[Sample[T]]:
    with open(path) as f:
        df = json.load(f)
    samples = [to_sample(x, sample_type) for x in df]
    return [s for s in samples if s.annotations]



from dataclasses import fields
from functools import partial
from operator import itemgetter
from typing import Generic, Protocol, TypeVar

import numpy as np
import torch
from toolz.functoolz import compose
from torchvision.ops import nms

from detectinhos.batch import Batch, apply_eval, pad, un_batch
from detectinhos.sample import Annotation, Sample
from detectinhos.sublosses import WeightedLoss

T = TypeVar("T")


class HasBoxesAndClasses(Protocol, Generic[T]):
    scores: T
    boxes: T
    classes: T


def decode(
    pred: HasBoxesAndClasses[torch.Tensor],
    sublosses: HasBoxesAndClasses[WeightedLoss],
    priors: torch.Tensor,
    nms_threshold: float = 0.4,
    confidence_threshold: float = 0.5,
) -> HasBoxesAndClasses[torch.Tensor]:
    n_batches = pred.boxes.shape[0]
    decoded_fields: dict[str, list[torch.Tensor]] = {
        f.name: [] for f in fields(sublosses)
    }

    for b in range(n_batches):
        # Decode boxes and scores
        boxes = sublosses.boxes.dec_pred(pred.boxes[b], priors)
        scores = sublosses.scores.dec_pred(pred.scores[b], priors)

        mask = scores > confidence_threshold
        keep = nms(boxes[mask], scores[mask], iou_threshold=nms_threshold)

        for field in fields(sublosses):
            name = field.name
            subloss = getattr(sublosses, name)
            raw = getattr(pred, name)[b]
            decoded = subloss.dec_pred(raw, priors)
            filtered = decoded[mask][keep]
            decoded_fields[name].append(filtered)

    output_cls = type(pred)
    return output_cls(**{k: pad(v) for k, v in decoded_fields.items()})


def infer_on_rgb(
    image: np.ndarray,
    model: torch.nn.Module,
    to_sample,
    decode,
    file: str = "",
):
    def to_batch(image, file="fake.png") -> Batch:
        return Batch(
            files=[file],
            image=torch.from_numpy(image)
            .permute(2, 0, 1)
            .float()
            .unsqueeze(0),
        )

    # On RGB
    sample = compose(
        compose(
            to_sample,
            itemgetter(0),
            un_batch,
            decode,
        ),
        partial(apply_eval, model=model),
        to_batch,
    )(image)
    sample.file_name = file
    return sample


def true2sample(
    true: HasBoxesAndClasses,
    to_sample,
) -> list[Sample[Annotation]]:
    return list(map(to_sample, un_batch(true)))



from typing import List, Tuple

import numpy as np
from mean_average_precision import MetricBuilder

from detectinhos.sample import Annotation, Sample


# TODO: For some reason this is imortant, fix me
def to_bbox(bbox):
    return bbox[0] * 640, bbox[1] * 480, bbox[2] * 640, bbox[3] * 480


def to_table(
    true: List[Sample[Annotation]],
    pred: List[Sample[Annotation]],
    mapping: dict[str, int],
) -> List[Tuple[np.ndarray, np.ndarray]]:
    total = []

    for strue, spred in zip(true, pred):
        # Format predictions
        pred_arr = np.array(
            [
                list(to_bbox(ann.bbox)) + [mapping[ann.label] - 1, ann.score]
                for ann in spred.annotations
            ],
            dtype=np.float32,
        )

        # Format ground truth
        true_arr = np.zeros((len(strue.annotations), 7), dtype=np.float32)
        for i, ann in enumerate(strue.annotations):
            true_arr[i, :4] = to_bbox(ann.bbox)
            # shift to zero-indexed class
            true_arr[i, 4] = mapping[ann.label] - 1

        total.append((pred_arr, true_arr))

    return total


class MeanAveragePrecision:
    def __init__(self, num_classes: int, mapping: dict[str, int]):
        # Convention: skip background class (index 0)
        self.metric_fn = MetricBuilder.build_evaluation_metric(
            "map_2d",
            async_mode=False,
            num_classes=num_classes - 1,
        )
        self.mapping = mapping

    def add(self, true: List[Sample], pred: List[Sample]) -> None:
        for pred_arr, true_arr in to_table(true, pred, mapping=self.mapping):
            self.metric_fn.add(pred_arr, true_arr)

    def value(self, iou_thresholds: float = 0.5) -> dict[str, float]:
        return self.metric_fn.value(iou_thresholds=iou_thresholds)

    def reset(self) -> None:
        self.metric_fn.reset()



import pytest

from detectinhos.sample import Annotation, Sample, read_dataset


@pytest.mark.parametrize(
    "resolution",
    [
        (480, 640),
    ],
)
def test_reads_dataset(annotations):
    dataset = read_dataset(annotations, sample_type=Sample[Annotation])
    assert len(dataset) > 0
    assert len(dataset[0].annotations) > 0



from typing import List, Tuple, Union

import torch


def point_form(boxes: torch.Tensor) -> torch.Tensor:
    return torch.cat(
        (
            boxes[..., :2] - boxes[..., 2:] / 2,
            boxes[..., :2] + boxes[..., 2:] / 2,
        ),
        dim=-1,
    )


def encode(
    boxes: torch.Tensor,
    priors: torch.Tensor,
    variances: List[float],
) -> torch.Tensor:
    # dist b/t match center and prior's center
    g_cxcy = (boxes[:, :2] + boxes[:, 2:]) / 2 - priors[..., :2]
    # encode variance
    g_cxcy /= variances[0] * priors[..., 2:]
    # match wh / prior wh
    g_wh = (boxes[:, 2:] - boxes[:, :2]) / priors[..., 2:]
    g_wh = torch.log(g_wh) / variances[1]
    # return target for smooth_l1_loss
    return torch.cat([g_cxcy, g_wh], 1)  # [num_priors,4]


def decode(
    boxes: torch.Tensor,
    priors: torch.Tensor,
    variances: Union[List[float], Tuple[float, float]],
) -> torch.Tensor:
    boxes = torch.cat(
        (
            priors[..., :2] + boxes[..., :2] * variances[0] * priors[..., 2:],
            priors[..., 2:] * torch.exp(boxes[..., 2:] * variances[1]),
        ),
        dim=-1,
    )
    boxes[..., :2] -= boxes[..., 2:] / 2
    boxes[..., 2:] += boxes[..., :2]

    return boxes


def to_shape(x, matched) -> torch.Tensor:
    return x.unsqueeze(1).expand(matched.shape[0], 5).unsqueeze(2)


def encode_landmarks(
    landmarks: torch.Tensor,
    priors: torch.Tensor,
    variances: Union[List[float], Tuple[float, float]],
) -> torch.Tensor:
    # dist b/t match center and prior's center

    # Reshape matched to [num_priors, 5, 2]
    landmarks = torch.reshape(landmarks, (landmarks.shape[0], 5, 2))

    # Expand priors to match the shape of matched using broadcasting
    # Change shape from [num_priors, 4] to [num_priors, 1, 4]

    priors = priors[:, None]

    # Calculate the distance between the match center and the prior's center
    g_cxcy = landmarks[:, :, :2] - priors[:, :, :2]

    # Encode variance
    g_cxcy /= variances[0] * priors[:, :, 2:]

    # Return target for smooth_l1_loss
    return g_cxcy.reshape(-1, 10)


def decode_landmarks(
    landmarks: torch.Tensor,
    priors: torch.Tensor,
    variances: Union[List[float], Tuple[float, float]],
) -> torch.Tensor:
    x = landmarks
    return torch.cat(
        (
            priors[..., :2] + x[..., :2] * variances[0] * priors[..., 2:],
            priors[..., :2] + x[..., 2:4] * variances[0] * priors[..., 2:],
            priors[..., :2] + x[..., 4:6] * variances[0] * priors[..., 2:],
            priors[..., :2] + x[..., 6:8] * variances[0] * priors[..., 2:],
            priors[..., :2] + x[..., 8:10] * variances[0] * priors[..., 2:],
        ),
        dim=-1,
    )



import pytest
import torch

from detectinhos.encode import (
    decode,
    decode_landmarks,
    encode,
    encode_landmarks,
)


@pytest.fixture
def matched(encode):
    if encode == encode_landmarks:
        # Landmarks ~ [batch_size, 10]
        return torch.arange(20).reshape(2, 10).float()
    # Boxes ~ [batch_size, 4]
    return torch.tensor([[1, 1, 3, 3], [2, 2, 4, 4.0]])


@pytest.fixture
def priors(encode):
    return torch.tensor([[0, 0, 2, 2], [1, 1, 3, 3.0]])


@pytest.fixture
def variances():
    return [0.1, 0.2]


@pytest.fixture
def encoded(encode, matched, priors, variances):
    # Call the encode function
    return encode(matched, priors, variances)


@pytest.fixture
def decoded(decode, encoded, priors, variances):
    # Call the decode function
    return decode(encoded, priors, variances)


@pytest.mark.parametrize(
    "encode, decode",
    [
        (encode, decode),
        (encode_landmarks, decode_landmarks),
    ],
)
def test_encodes_decodes(decoded, matched):
    # Assert the decoded matches the original matched data
    assert torch.allclose(decoded, matched, atol=1e-4)



from dataclasses import dataclass, fields
from typing import Generic, List, Optional, Protocol, Sequence, TypeVar

import numpy as np
import torch
from torch.nn.utils.rnn import pad_sequence

T = TypeVar("T")


class HasBoxesAndClasses(Protocol, Generic[T]):
    boxes: T
    classes: T
    scores: T


# A single element in the batch
@dataclass
class BatchElement(Generic[T]):
    file: str
    image: torch.Tensor
    true: HasBoxesAndClasses[T]


# Stacked BatchElements along batch dimension
@dataclass
class Batch(Generic[T]):
    files: list[str]
    image: T
    # Can be optional when we are doing inference
    true: Optional[HasBoxesAndClasses[T]] = None


def apply_eval(
    batch: Batch[T],
    model: torch.nn.Module,
) -> HasBoxesAndClasses[T]:
    original_mode = model.training
    model.eval()
    predicted = model(batch.image)
    model.train(original_mode)
    return predicted


def pad(tensors: Sequence[torch.Tensor]) -> torch.Tensor:
    return pad_sequence(
        tensors,
        batch_first=True,
        padding_value=float("nan"),
    )


def to_batch(
    x: Sequence[HasBoxesAndClasses[np.ndarray]],
) -> HasBoxesAndClasses[torch.Tensor]:
    if not x:
        raise ValueError("Received an empty sequence of data")

    cls = type(x[0])
    field_names = [f.name for f in fields(x[0])]

    batched = {
        name: pad([torch.Tensor(getattr(t, name)) for t in x])
        for name in field_names
    }
    return cls(**batched)


def detection_collate(
    batch: list[BatchElement],
) -> Batch:
    images = torch.stack([torch.Tensor(sample.image) for sample in batch])
    files = [sample.file for sample in batch]
    return Batch(files, images, to_batch([b.true for b in batch]))


def un_batch(
    x: HasBoxesAndClasses[torch.Tensor],
) -> List[HasBoxesAndClasses[np.ndarray]]:
    cls = type(x)
    fnames = [f.name for f in fields(x)]  # assumes x is a dataclass
    batch_size = x.boxes.shape[0]

    result: List[HasBoxesAndClasses[np.ndarray]] = []
    for i in range(batch_size):
        valid = ~torch.isnan(x.boxes[i]).any(dim=-1)
        sample = {
            name: getattr(x, name)[i][valid].cpu().numpy() for name in fnames
        }
        result.append(cls(**sample))
    return result



from functools import partial
from typing import Callable

import numpy as np
import pytest
import torch

from detectinhos.batch import detection_collate
from detectinhos.dataset import DetectionDataset
from detectinhos.inference import (
    decode as decode_generic,
    infer_on_rgb,
    true2sample,
)
from detectinhos.loss import DetectionLoss
from detectinhos.metrics import MeanAveragePrecision
from detectinhos.sample import Annotation, Sample, read_dataset
from detectinhos.tasks.standard import (
    TASK,
    DetectionTargets,
    build_targets,
)


def approx(x):
    return pytest.approx(x, abs=0.001)


class DedetectionModel(torch.nn.Module):
    def __init__(
        self,
        n_clases: int,
        classes: torch.Tensor,
        boxes: torch.Tensor,
    ) -> None:
        super().__init__()
        self.n_clases = n_clases
        self.classes = classes
        self.boxes = boxes

    def forward(self, images: torch.Tensor) -> DetectionTargets:
        # Expand classes_pred to shape [batch_size, n_anchors, n_clases]
        batch_size = images.shape[0]
        boxes = self.boxes.expand(batch_size, -1, 4).clone()
        classes = self.classes.expand(batch_size, -1, -1).clone()
        return DetectionTargets(
            # Return the same tensor twice, one for scores another for labels
            scores=classes,
            classes=classes,
            boxes=boxes,
        )


@pytest.fixture
def build_model(
    classes_pred,
    boxes_pred,
) -> Callable[[int], DedetectionModel]:
    def build_model(n_clases: int) -> DedetectionModel:
        return DedetectionModel(
            n_clases=n_clases,
            classes=classes_pred,
            boxes=boxes_pred,
        )

    return build_model


@pytest.mark.parametrize(
    "batch_size",
    [
        4,
    ],
)
@pytest.mark.parametrize(
    "resolution",
    [
        (480, 640),
    ],
)
def test_training_loop(
    batch_size,
    annotations,
    build_model,
    sample_anchors,
    resolution,
):
    mapping = {"background": 0, "apple": 1}
    to_sample, to_targets = build_targets(mapping)

    dataloader = torch.utils.data.DataLoader(
        DetectionDataset(
            labels=read_dataset(annotations, Sample[Annotation]) * 8,
            to_targets=to_targets,
        ),
        batch_size=batch_size,
        num_workers=1,
        collate_fn=detection_collate,
    )

    model = build_model(
        n_clases=2,
    )
    loss = DetectionLoss(
        priors=sample_anchors,
        sublosses=TASK,
    )
    decode = partial(
        decode_generic,
        sublosses=TASK,
        priors=sample_anchors,
        confidence_threshold=0.2,
        nms_threshold=1.0,
    )
    to_samples = partial(
        true2sample,
        to_sample=to_sample,
    )

    decode_image = partial(
        decode_generic,
        sublosses=TASK,
        priors=sample_anchors,
        confidence_threshold=0.5,
        nms_threshold=0.4,
    )

    to_samples = partial(
        true2sample,
        to_sample=to_sample,
    )

    # TODO: Fix this: providing mapping and num classes -- is redundant
    map_metric = MeanAveragePrecision(num_classes=2, mapping=mapping)

    # sourcery skip: no-loop-in-tests
    for batch in dataloader:
        batch.pred = model(batch.image)
        batch.true.classes = batch.true.classes.long()
        losses = loss(batch.pred, batch.true)
        map_metric.add(
            true=to_samples(batch.true),
            pred=to_samples(decode(batch.pred)),
        )
        # Test 1: Check forward pass and loss
        assert "loss" in losses

    # Test 2: Check mAP metric is calculated correctly
    assert map_metric.value()["mAP"] == pytest.approx(0.5)

    # Now check the inference after training
    infer_on_rgb_vanilla = partial(
        infer_on_rgb,
        model=model,
        to_sample=to_sample,
        decode=decode_image,
    )

    # Test 3: Now check the inference works
    sample = infer_on_rgb_vanilla(np.random.randint(0, 255, resolution + (3,)))
    assert len(sample.annotations) == 1
    assert sample.annotations[0].label == "apple"
    assert sample.annotations[0].score == approx(0.62)
    assert sample.annotations[0].bbox == approx([0.645, 0.813, 0.805, 0.956])



from dataclasses import dataclass
from functools import partial
from typing import Callable, Generic, TypeVar

import numpy as np
import torch

from detectinhos.encode import decode as decode_boxes, encode
from detectinhos.sample import Annotation, Sample
from detectinhos.sublosses import (
    WeightedLoss,
    masked_loss,
    retina_confidence_loss,
)

T = TypeVar(
    "T",
    np.ndarray,
    torch.Tensor,
    WeightedLoss,
)


# TODO: Add the pure vanilla tests
@dataclass
class DetectionTargets(Generic[T]):
    scores: T  # [B, N]
    boxes: T
    classes: T


def to_sample(
    predicted: DetectionTargets[np.ndarray],
    inverse_mapping: dict[int, str],
    file_name: str = "",
) -> Sample:
    predictions = zip(
        predicted.boxes.tolist(),
        predicted.classes.reshape(-1).tolist(),
        predicted.scores.reshape(-1).tolist(),
    )
    return Sample(
        file_name=file_name,
        annotations=[
            Annotation(
                bbox=box,
                label=inverse_mapping[label],
                score=score,
            )
            for box, label, score in predictions
        ],
    )


def to_targets(
    sample: Sample,
    mapping: dict[str, int],
) -> DetectionTargets[np.ndarray]:
    bboxes = []
    label_ids = []
    scores = []

    for label in sample.annotations:
        bboxes.append(label.bbox)

        label_id = mapping.get(label.label, 0)
        label_ids.append([label_id])
        scores.append([label.score])

    return DetectionTargets(
        boxes=np.array(bboxes),
        classes=np.array(label_ids, dtype=np.int64),
        scores=np.array(scores, dtype=np.float32),
    )


def build_targets(
    mapping: dict[int, str],
) -> tuple[
    Callable[[DetectionTargets[np.ndarray]], Sample],
    Callable[[Sample], DetectionTargets[np.ndarray]],
]:
    inverse_mapping = {v: k for k, v in mapping.items()}
    return (
        partial(to_sample, inverse_mapping=inverse_mapping),
        partial(to_targets, mapping=mapping),
    )


TASK = DetectionTargets(
    scores=WeightedLoss(
        loss=None,
        # NB: drop the background class
        dec_pred=lambda logits, _: torch.nn.functional.softmax(logits, dim=-1)[
            ..., 1:
        ].max(dim=-1)[0],
    ),
    classes=WeightedLoss(
        loss=retina_confidence_loss,
        weight=2.0,
        enc_pred=lambda x, _: x.reshape(-1, 2),
        enc_true=lambda x, _: x,
        # NB: drop the background class, labels += 1
        dec_pred=lambda logits, _: (
            torch.nn.functional.softmax(logits, dim=-1)[..., 1:].max(dim=-1)[1]
            + 1
        ).float(),
        needs_negatives=True,
    ),
    boxes=WeightedLoss(
        loss=masked_loss(torch.nn.SmoothL1Loss()),
        weight=1.0,
        enc_pred=lambda x, _: x,
        enc_true=partial(encode, variances=[0.1, 0.2]),
        dec_pred=partial(decode_boxes, variances=[0.1, 0.2]),
        needs_negatives=False,
    ),
)



from pathlib import Path
from typing import Callable, TypeVar

import cv2
import numpy as np
import torch

from detectinhos.batch import BatchElement
from detectinhos.sample import Sample

T = TypeVar(
    "T",
    np.ndarray,
    torch.Tensor,
)


def do_nothing(x: np.ndarray, y: T) -> tuple[np.ndarray, T]:
    return x, y


Augmentation = Callable[[np.ndarray, T], tuple[np.ndarray, T]]


def load_rgb(image_path: Path | str) -> np.ndarray:
    image = cv2.imread(str(image_path))
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    return image


class DetectionDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        labels: list[Sample],
        to_targets: Callable[[Sample], T],
        transform: Augmentation = do_nothing,
    ) -> None:
        self.transform = transform
        self.labels = labels
        self.to_targets = to_targets

    def __len__(self) -> int:
        return len(self.labels)

    def __getitem__(self, index: int) -> BatchElement[torch.Tensor]:
        sample = self.labels[index]
        raw_image = load_rgb(sample.file_name)
        raw_targets = self.to_targets(sample)
        image, targets = self.transform(
            raw_image,
            raw_targets,
        )
        return BatchElement(
            file=sample.file_name,
            image=image,
            true=targets,
        )
